(in-package "ACL2")

; All theorems in this file are local except for those that are to be exported,
; which are labeled with the comment "; Exported theorem:".  We use def-gl-thm
; (actually a local version) so dispatch lemmas that were seen to be needed
; upon analysis of simplification checkpoints.  Without def-gl-thm, we would
; consider using one of these libraries:

; (local (include-book "arithmetic-5/top" :dir :system))
; (local (include-book "rtl/rel8/lib/top" :dir :system))

; That would perhaps take considerable work, however, because wm08 uses lognot,
; which doesn't appear in many lemmas in the above libraries.  Less important
; is that we might have to disable some rules, e.g. ash-rewrite in the latter
; case.  Fortunately, finding lemmas in either of the above books isn't hard,
; for example using the find-lemmas utility:

; (local (include-book "misc/find-lemmas" :dir :system))

(include-book "x86-state")

; Exported theorem:
(defthm rgfi-update-rgfi
  (equal (rgfi i (update-rgfi i v x86-32))
         v)
  :hints (("Goal" :in-theory (enable rgfi update-rgfi))))

; Exported theorem:
(defthm rgfi-read-through-different-address-update-rgfi
  (implies (and (n03p i)
                (n03p j)
                (not (equal i j)))
           (equal (rgfi i (update-rgfi j v x86-32))
                  (rgfi i x86-32)))
  :hints (("Goal" :in-theory (enable rgfi update-rgfi))))

; Exported theorem:
(defthm memi-update-memi
  (equal (memi i (update-memi i v x86-32))
         v)
  :hints (("Goal" :in-theory (enable memi update-memi))))

; Exported theorem:
(defthm memi-read-through-different-address-update-memi
  (implies (and (n32p i)
                (n32p j)
                (not (equal i j)))
           (equal (memi i (update-memi j v x86-32))
                  (memi i x86-32)))
  :hints (("Goal" :in-theory (enable memi update-memi))))

; Comment from Warren:
; Below are conjectures about the functions I use to access the
; memory.  I haven't yet introduced 16-bit versions, but I will need
; to do so for X86 model.  The Y86 doesn't have 16-bit operations.

(local (include-book "centaur/gl/gl" :dir :system))

(defmacro def-local-gl-thm (&rest args)
  `(local (def-gl-thm ,@args)))

(def-local-gl-thm rm08-wm08-lemma-1
  :hyp (force (and (integerp v)
                   (<= 0 v)
                   (< v 256)
                   (n32p mem-val)
                   (integerp i02)
                   (<= 0 i02)
                   (<= i02 3)))
  :concl (equal
          (logand (ash (logior (logand (+ -1 (- (ash 255 (ash i02 3))))
                                       mem-val)
                               (ash v (ash i02 3)))
                       (- (ash i02 3)))
                  255)
          v)
  :g-bindings
  '((mem-val
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))
    (v
     (:g-number (33 34 35 36 37 38 39 40 41)))
    (i02
     (:g-number (42 43 44)))))

; From x86-state.lisp:
(def-local-gl-thm ash-addr--2-is-less
    :hyp (n32p addr)
    :concl (< (ash addr -2) 1073741824)
    :rule-classes :linear
    :g-bindings
    '((addr
       (:g-number ( 0  1  2  3  4  5  6  7  8  9  10 11 12 13 14 15 16
                       17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))))

; Exported theorem:
(defthm rm08-wm08
  (implies (and (x86-32p x86-32)
                (n32p i)
                (n08p v))
           (equal (rm08 i (wm08 i v x86-32))
                  v))
  :hints (("Goal" :in-theory (enable rm08 wm08))))

(def-local-gl-thm rm08-read-through-other-address-wm08-same-dword-lemma
  :hyp (force (and (n32p mem-val)
                   (integerp i02)
                   (<= 0 i02)
                   (< i02 4)
                   (integerp j02)
                   (<= 0 j02)
                   (< j02 4)
                   (not (equal i02 j02))
                   (integerp v)
                   (<= 0 v)
                   (< v 256)))
  :concl (equal
          (logand (ash (logior (logand (+ -1 (- (ash 255 (ash j02 3))))
                                       mem-val)
                               (ash v (ash j02 3)))
                       (- (ash i02 3)))
                  255)
          (logand (ash mem-val
                       (- (ash i02 3)))
                  255))
  :g-bindings
  '((mem-val
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))
    (i02
     (:g-number (33 34 35 36 37)))
    (j02
     (:g-number (38 39 40 41 42)))
    (v
     (:g-number (43 44 45 46 47 48 49 50 51)))))

(def-local-gl-thm n32p-concat-ash-logand
  :hyp (n32p i)
  :concl (equal (logior (ash (ash i -2) 2)
                        (logand i 3))
                i)
  :rule-classes nil
  :g-bindings
  '((i
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))))

(local
 (defthm rm08-read-through-other-address-wm08-same-dword
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (equal (ash i -2) (ash j -2))
                 (not (equal i j))
                 (n08p v))
            (equal (rm08 i (wm08 j v x86-32))
                   (rm08 i x86-32)))
   :hints (("Goal"
            :in-theory (enable rm08 wm08))
           ("[1]Goal''"
            :use ((:instance n32p-concat-ash-logand (i i))
                  (:instance n32p-concat-ash-logand (i j)))))
   :rule-classes nil))

(local (in-theory (disable
                   rm08-read-through-other-address-wm08-same-dword-lemma)))

(local
 (defthm rm08-read-through-other-address-wm08-different-dword
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (not (equal (ash i -2) (ash j -2)))
                 (not (equal i j))
                 (n08p v))
            (equal (rm08 i (wm08 j v x86-32))
                   (rm08 i x86-32)))
   :hints (("Goal" :in-theory (enable rm08 wm08)))
   :rule-classes nil))

; Exported theorem:
(defthm rm08-read-through-other-address-wm08
  (implies (and (x86-32p x86-32)
                (n32p i)
                (n32p j)
                (not (equal i j))
                (n08p v))
           (equal (rm08 i (wm08 j v x86-32))
                  (rm08 i x86-32)))
  :hints (("Goal" :use (rm08-read-through-other-address-wm08-different-dword
                        rm08-read-through-other-address-wm08-same-dword))))

; Start proof of rm32-wm32.

(def-local-gl-thm rm32-wm32-lemma-1
  :hyp (n32p v)
  :concl (equal (logior (ash (logand (ash v -24) 255) 24)
                        (ash (logand (ash v -16) 255) 16)
                        (ash (logand (ash v -8) 255) 8)
                        (logand v 255))
                v)
  :g-bindings
  '((v
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))))

(def-local-gl-thm rm32-wm32-lemma-2
  :hyp (and (n32p i)
            (integerp k1)
            (<= 0 k1)
            (< k1 4)
            (integerp k2)
            (<= 0 k2)
            (< k2 4))
  :concl (equal (equal (logand (+ k1 i) 4294967295)
                       (logand (+ k2 i) 4294967295))
                (equal k1 k2))
  :g-bindings
  '((i
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))
    (k1
     (:g-number (33 34 35)))
    (k2
     (:g-number (36 37 38)))))

(def-local-gl-thm rm32-wm32-lemma-3
  :hyp (and (n32p i)
            (integerp k2)
            (<= 0 k2)
            (< k2 4))
  :concl (equal (equal i
                       (logand (+ k2 i) 4294967295))
                (equal 0 k2))
  :g-bindings
  '((i
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))
    (k2
     (:g-number (33 34 35)))))

; Exported theorem:
(defthm rm32-wm32
  (implies (and (x86-32p x86-32)
                (n32p i)
                (n32p v))
           (equal (rm32 i (wm32 i v x86-32))
                  v))
  :hints (("Goal" :in-theory (enable rm32 wm32))))

; Start proof of rm32-read-through-other-address-wm32.  That theorem really
; consists of two cases: one for i<j and the other for j<i.  See the comment in
; that theorem.  We prove each case separately.

; Start proof of rm32-read-through-other-address-wm32-i<j.

(def-local-gl-thm rm32-read-through-other-address-wm32-i<j-lemma-1
  :hyp (and (n32p i)
            (n32p j)
            (<= (+ 4 i) j)
            (<= (+ 4 j) 4294967296) ; (expt 2 32)
            (integerp k1)
            (<= 0 k1)
            (< k1 4)
            (integerp k2)
            (<= 0 k2)
            (< k2 4))
  :concl (equal (equal (logand (+ k1 i) 4294967295)
                       (logand (+ k2 j) 4294967295))
                nil)
  :g-bindings
  '((i
     (:g-number (0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42
                   44 46 48 50 52 54 56 58 60 62 64)))
    (j
     (:g-number (1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43
                   45 47 49 51 53 55 57 59 61 63 65)))
    (k1
     (:g-number (66 67 68)))
    (k2
     (:g-number (69 70 71)))))

(def-local-gl-thm rm32-read-through-other-address-wm32-i<j-lemma-2
  :hyp (and (n32p i)
            (n32p j)
            (<= (+ 4 i) j))
  :concl (equal (equal (ash i -2)
                       (ash j -2))
                nil)
  :g-bindings
  '((i
     (:g-number (0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42
                   44 46 48 50 52 54 56 58 60 62 64)))
    (j
     (:g-number (1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43
                   45 47 49 51 53 55 57 59 61 63 65)))))

(local
 (defthm memi-read-through-other-address-wm08-i<j
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (<= (+ 4 i) j)
                 (n08p v))
            (equal (memi (ash i -2) (wm08 j v x86-32))
                   (memi (ash i -2) x86-32)))
   :hints (("Goal" :in-theory (enable wm08)))))

(def-local-gl-thm rm32-read-through-other-address-wm32-i<j-lemma-3
  :hyp (and (n32p j)
            (<= (+ 4 j) 4294967296) ; (expt 2 32)
            (n02p k))
  :concl (equal (logand (+ k j) 4294967295) ; (1- (expt 2 32))
                (+ k j))
  :g-bindings
  '((j
     (:g-number ( 0  1  2  3  4  5  6  7  8  9 10 11 12 13 14 15 16
                     17 18 19 20 21 22 23 24 25 26 27 28 29 30 31 32)))
    (k
     (:g-number (33 34 35)))))

(def-local-gl-thm rm32-read-through-other-address-wm32-i<j-lemma-4
  :hyp (and (n32p i)
            (n32p j)
            (< i j)
            (equal (logand j 3) 0))
  :concl (equal (equal (ash i -2)
                       (ash j -2))
                nil)
  :g-bindings
  '((i
     (:g-number (0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42
                   44 46 48 50 52 54 56 58 60 62 64)))
    (j
     (:g-number (1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43
                   45 47 49 51 53 55 57 59 61 63 65)))))

(local
 (defthm rm08-update-memi
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (< i j)
                 (equal (logand j 3) 0))
            (equal (rm08 i
                         (update-memi (ash j -2) v x86-32))
                   (rm08 i x86-32)))
   :hints (("Goal" :in-theory (enable rm08)))))

(local
 (defthm rm32-read-through-other-address-wm32-i<j
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 ;; Note that (not (= i j)) is not strong enough:
                 (<= (+ 4 i) j)
                 (<= (+ 4 j) 4294967296) ; (expt 2 32)
                 )
            (equal (rm32 i (wm32 j v x86-32))
                   (rm32 i x86-32)))
   :hints (("Goal" :in-theory (enable rm32 wm32)))
   :rule-classes nil))

; Start proof of rm32-read-through-other-address-wm32-j<i.

(def-local-gl-thm rm32-read-through-other-address-wm32-j<i-lemma-1
  :hyp (and (n32p i)
            (n32p j)
            (< j i)
            (equal (logand i 3) 0))
  :concl (equal (equal (ash i -2)
                       (ash j -2))
                nil)
  :g-bindings
  '((i
     (:g-number (0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42
                   44 46 48 50 52 54 56 58 60 62 64)))
    (j
     (:g-number (1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43
                   45 47 49 51 53 55 57 59 61 63 65)))))

(local
 (defthm memi-read-through-other-address-j<i-wm08
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (< j i)
                 (EQUAL (LOGAND I 3) 0)
                 (n08p v))
            (equal (memi (ash i -2) (wm08 j v x86-32))
                   (memi (ash i -2) x86-32)))
   :hints (("Goal" :in-theory (enable wm08)))))

(def-local-gl-thm rm32-read-through-other-address-wm32-j<i-lemma-2
  :hyp (and (n32p i)
            (n32p j)
            (<= (+ 4 j) i))
  :concl (equal (equal (ash i -2)
                       (ash j -2))
                nil)
  :g-bindings
  '((i
     (:g-number (0 2 4 6 8 10 12 14 16 18 20 22 24 26 28 30 32 34 36 38 40 42
                   44 46 48 50 52 54 56 58 60 62 64)))
    (j
     (:g-number (1 3 5 7 9 11 13 15 17 19 21 23 25 27 29 31 33 35 37 39 41 43
                   45 47 49 51 53 55 57 59 61 63 65)))))

(local
 (defthm rm08-update-memi-j<i
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 (<= (+ 4 j) i)
                 (equal (logand j 3) 0))
            (equal (rm08 i
                         (update-memi (ash j -2) v x86-32))
                   (rm08 i x86-32)))
   :hints (("Goal" :in-theory (enable rm08)))))

(local
 (defthm rm32-read-through-other-address-wm32-j<i
   (implies (and (x86-32p x86-32)
                 (n32p i)
                 (n32p j)
                 ;; Note that (not (= i j)) is not strong enough:
                 (<= (+ 4 j) i)
                 (<= (+ 4 i) 4294967296) ; (expt 2 32)
                 )
            (equal (rm32 i (wm32 j v x86-32))
                   (rm32 i x86-32)))
   :hints (("Goal" :in-theory (enable rm32 wm32)))
   :rule-classes nil))

; Exported theorem:
(defthm rm32-read-through-other-address-wm32
  (implies (and (x86-32p x86-32)
                (n32p i)
                (n32p j)

; Note that (not (= i j)) is not a strong enough hypothesis, because the dwords
; starting at i and at j might overlap.  Moreover, we can't tolerate
; wrap-around: for example, for the first case below, if j is 2^32-1 and i is
; 0, then (<= (+ 4 i) j) holds but the dwords overlap because of wm32 will wrap
; around; hence the conjunct (<= (+ 4 j) 2^32).

                (or (and (<= (+ 4 i) j)
                         (<= (+ 4 j) 4294967296)) ; (expt 2 32) ;
                    (and (<= (+ 4 j) i)
                         (<= (+ 4 i) 4294967296))))
           (equal (rm32 i (wm32 j v x86-32))
                  (rm32 i x86-32)))
  :hints (("Goal" :use (rm32-read-through-other-address-wm32-i<j
                        rm32-read-through-other-address-wm32-j<i))))

; Comment from Warren:
; Of course, when these are proven to be theorems, then combinations
; of such accesses will be desired.  For instance, what happens if a
; 32-bit number is partially written to by wm08 -- and so on...

